use crate::instructions::{get_instruction, Instruction};
use crate::memory::Mem;
use crate::types::{Byte, Word};
use std::cell::RefCell;
use std::path::PathBuf;
use std::rc::Rc;
use std::str::FromStr;
use std::sync::mpsc::Receiver;
use std::sync::Arc;
use std::sync::Mutex;
use std::thread::sleep;
use std::time::Duration;

use anyhow::{bail, Result};

#[derive(Clone, Copy)]
pub enum StatusFlag {
    Negative = 0b1000_0000,
    Overflow = 0b0100_0000,
    Brk = 0b0011_0000,
    //BrkIrq = 0b0010_0000,
    Decimal = 0b0000_1000,
    IrqDisable = 0b0000_0100,
    Zero = 0b000_0010,
    Carry = 0b0000_0001,
}

pub struct Cpu {
    pub a: Byte,  // Accumulator Register
    pub x: Byte,  // X Register
    pub y: Byte,  // Y Register
    pub pc: Word, // Program Counter
    pub s: Byte,  // Stack Pointer
    pub p: Byte,  // Status Register
    pub irq: Receiver<bool>,
    pub nmi: bool,
    pub memory: Arc<Mutex<Mem>>,
    pub pending_cycles: usize,
    cycle_count: usize,
}

impl Cpu {
    pub fn new(memory: Arc<Mutex<Mem>>, irq: Receiver<bool>) -> Self {
        Cpu {
            a: 0x00,
            x: 0x00,
            y: 0x00,
            pc: 0x0000,
            s: 0xFF,
            p: 0b0010_0100,
            irq,
            nmi: false,
            memory,
            pending_cycles: 0,
            cycle_count: 0,
        }
    }
    pub fn reset(&mut self) -> Result<()> {
        let reset_vector_pointer = self.read_word(0xFFFC)?;
        self.pc = reset_vector_pointer;
        self.pending_cycles = 0;
        Ok(())
    }
    pub fn read(&self, address: Word) -> Result<Byte> {
        let memory = match self.memory.lock() {
            Ok(read) => read,
            Err(_) => {
                bail!("Couldn't acquire lock on memory in cpu thread")
            }
        };
        Ok(memory.read(address))
    }
    pub fn read_word(&self, address: Word) -> Result<Word> {
        let low_byte = self.read(address)?;
        let high_byte = self.read(address + 0x1)?;
        Ok((high_byte as u16) << 8 | (low_byte as u16))
    }

    fn stack_addr(&self) -> Word {
        // Dunno if this is necessary, i just don't like adding the 0x0100 every time
        0x0100 + self.s as u16
    }

    pub fn push_stack(&mut self, data: Byte) -> Result<()> {
        self.s = self.s.wrapping_sub(0x1);
        self.write(self.stack_addr(), data)?;
        Ok(())
    }

    pub fn push_stack_word(&mut self, address: Word) -> Result<()> {
        self.s = self.s.wrapping_sub(0x1);
        self.write(self.stack_addr(), address.to_le_bytes()[1])?; // Upper byte first
        self.s = self.s.wrapping_sub(0x1);
        self.write(self.stack_addr(), address.to_le_bytes()[0])?; // Lower byte second
        Ok(())
    }

    pub fn pop_stack(&mut self) -> Result<Byte> {
        let byte = self.read(self.stack_addr())?;
        self.s = self.s.wrapping_add(0x1);
        Ok(byte)
    }

    pub fn pop_stack_word(&mut self) -> Result<Word> {
        let low_byte = self.pop_stack()?;
        let high_byte = self.pop_stack()?;
        let word = ((high_byte as Word) << 8) + low_byte as u16;
        Ok(word)
    }

    pub fn set_flag(&mut self, flag: StatusFlag, value: bool) {
        self.p &= !(flag as Byte);
        if value {
            self.p |= flag as Byte
        }
    }

    pub fn get_flag(&self, flag: StatusFlag) -> bool {
        (self.p & flag as Byte) > 0
    }

    pub fn is_negative(&self, value: Byte) -> bool {
        value & 0b1000_0000 == 0b1000_0000
    }

    pub fn write(&mut self, address: Word, data: Byte) -> Result<()> {
        let mut memory = match self.memory.lock() {
            Ok(write) => write,
            Err(_) => {
                bail!("Couldn't acquire write lock on memory in cpu thread")
            }
        };
        memory.write(address, data);
        Ok(())
    }

    pub fn execute(&mut self) {
        self.cycle();
        while self.pending_cycles != 0 {
            self.cycle();
        }
    }

    pub fn wait_for_interrupt(&self) {
        unimplemented!()
    }

    pub fn interrupt(&mut self) {
        self.push_stack_word(self.pc).unwrap();
        self.push_stack(self.p).unwrap();
        self.set_flag(StatusFlag::IrqDisable, true);
        self.pc = self.read_word(0xFFFE).unwrap();
    }

    pub fn cycle(&mut self) {
        while self.pending_cycles != 0 {
            // roughly cycle-accurate timing
            sleep(Duration::from_nanos(100));
            self.pending_cycles -= 1;
        }
        // if !self.get_flag(StatusFlag::IrqDisable) && self.irq.recv().unwrap() {
        //     self.interrupt();
        // }
        let opcode = match self.read(self.pc) {
            Ok(byte) => byte,
            Err(_) => {
                println!("Failed to read from memory at address {:#06x}!", self.pc);
                return;
            }
        };
        let instruction = get_instruction(opcode);
        match instruction {
            Instruction::Valid(valid_instruction) => {
                // println!("a: {a:#04x}, x: {x:#04x}, y: {y:#04x}, pc: {pc:#06x}, sp: {s:#04x}, sr: {p:#010b}, irq: {irq:?}, nmi: {nmi:?}", a = self.a, x = self.x, y = self.y, pc = self.pc, s = self.s, p = self.p, irq = self.irq, nmi = self.nmi);
                // println!(
                //     "Instruction: {:?}, {:#04x}",
                //     valid_instruction.opcode, opcode
                // );
                // println!("");
                self.pc += 1;
                match valid_instruction.opcode.call(self) {
                    Ok(_) => {
                        self.pending_cycles += valid_instruction.cycles as usize;
                    }
                    Err(_) => {
                        println!("An IncompatibleAddrMode was used at address {:#06x} for instruction {:?}", self.pc, valid_instruction);
                    }
                };
            }
            Instruction::Invalid(invalid_instruction) => match invalid_instruction.opcode {
                0x02 => {
                    let memory = match self.memory.lock() {
                        Ok(read) => read,
                        Err(_) => {
                            println!("Couldn't acquire read lock on memory in cpu thread");
                            return;
                        }
                    };

                    // println!("a: {a:#04x}, x: {x:#04x}, y: {y:#04x}, pc: {pc:#06x}, sp: {s:#04x}, sr: {p:#010b}, irq: {irq:?}, nmi: {nmi:?}", a = self.a, x = self.x, y = self.y, pc = self.pc, s = self.s, p = self.p, irq = self.irq.try_recv().unwrap_or_default(), nmi = self.nmi);
                    memory
                        .dump(PathBuf::from_str("./cpu_dump.bin").unwrap())
                        .unwrap();
                }
                _ => {
                    println!(
                        "An invalid instruction with opcode {:#04x} was called at address {:#06x}",
                        invalid_instruction.opcode, self.pc
                    );
                }
            },
        }
        self.cycle_count += 1;
    }

    pub fn stop(&mut self) {
        unimplemented!()
    }
}
