// #![allow(dead_code)]

mod cpu;
mod error;
mod instructions;
mod keyboard;
mod memory;
mod tui;
mod types;
mod video;

use crate::cpu::Cpu;
use crate::keyboard::Keyboard;
use crate::memory::Mem;
use crate::video::Crtc;

// use clap::Parser;
use minifb::{Scale, ScaleMode, Window, WindowOptions};
use serde::{Deserialize, Serialize};

use std::{
    fs::File,
    io::{stdout, Read, Result},
    path::PathBuf,
    str::FromStr,
    sync::{mpsc, Arc, Mutex},
    thread,
};

use crossterm::{
    event::{self, KeyCode, KeyEventKind},
    terminal::{disable_raw_mode, enable_raw_mode, EnterAlternateScreen, LeaveAlternateScreen},
    ExecutableCommand,
};

use ratatui::prelude::{CrosstermBackend, Terminal};

// use toml::Table;

//#[derive(Parser)]
//struct Cli {
//    // Load a rom onto the system rom
//    #[arg(short, required = true)]
//    rom: Option<std::path::PathBuf>,
//}

#[derive(Serialize, Deserialize, Debug)]
struct Config {
    char_rom: Option<String>,
    rom: String,
}

fn main() -> Result<()> {
    stdout().execute(EnterAlternateScreen)?;
    enable_raw_mode()?;
    let mut terminal = Terminal::new(CrosstermBackend::new(stdout()))?;
    terminal.clear()?;

    let config: Config = match File::open("./config.toml") {
        Ok(mut file) => {
            let mut string = String::new();
            file.read_to_string(&mut string).unwrap();
            toml::from_str(string.as_str()).unwrap()
        }
        Err(_) => return Ok(()),
    };

    let mut memory = Mem::new();
    let rom = match std::fs::File::open(config.rom) {
        Ok(file) => file,
        Err(error) => panic!("Couldn't open main rom! {:?}", error),
    };
    if let Err(error) = memory.load_rom(rom) {
        println!("{:?}", error);
    };

    memory
        .dump(PathBuf::from_str("./coredump.bin").unwrap())
        .unwrap();

    let shared_memory = Arc::new(Mutex::new(memory));

    let cpu_memory = shared_memory.clone();
    let display_memory = shared_memory.clone();
    let keyboard_memory = shared_memory.clone();

    let (interrupt_tx, interrupt_rx) = mpsc::channel();
    let (window_tx, window_rx) = mpsc::channel();

    thread::spawn(move || {
        let mut screen = Crtc::new(
            display_memory,
            config.char_rom.as_ref(),
            interrupt_tx,
            window_tx,
        );
        screen.run();
    });

    let mut window = Window::new(
        "ʕ·ᴥ·ʔ-☆",
        512,
        380,
        WindowOptions {
            resize: true,
            borderless: true,
            title: true,
            transparency: false,
            scale: Scale::X2,
            scale_mode: ScaleMode::AspectRatioStretch,
            topmost: false,
            none: true,
        },
    )
    .unwrap();

    window.set_input_callback(Box::new(Keyboard::new(keyboard_memory)));

    let cpu = Cpu::new(cpu_memory, interrupt_rx);
    let mut tui = tui::App::new(cpu);
    tui.init();

    loop {
        if event::poll(std::time::Duration::from_millis(16))? {
            if let event::Event::Key(key) = event::read()? {
                if key.kind == KeyEventKind::Press && key.code == KeyCode::Char('q') {
                    break;
                }
            }
        }
        let buffer = window_rx.recv().unwrap();
        window.update_with_buffer(&buffer, 512, 380).unwrap();
        tui.update(&mut terminal)?;
    }
    stdout().execute(LeaveAlternateScreen)?;
    disable_raw_mode()?;
    Ok(())
}
